import ephyviewer

import numpy as np
from  ephyviewer.tests.testing_tools import make_fake_signals


def test_spectrogramviewer(interactive=False):
    
    
    app = ephyviewer.mkQApp()
    
    fs = sample_rate = 100000.
    times = np.arange(0, 100., 1./sample_rate, dtype='float64')
    sigs = np.random.randn(times.size, 2) * 2.
    sigs[:, 0] += np.sin(times * 2 * np.pi *5000.) * 1.5
    sigs[:, 1] += np.sin(times * 2 * np.pi *20000.) * 0.5


    #create moving sinus
    sigsize = times.size
    f1 = 1000.
    f2 = 25000.
    freqs1 = np.concatenate([np.linspace(0,.5,sigsize*3//8) *  (f2-f1) + f1, np.ones(sigsize//4)*(f1+f2)/2 , np.linspace(0,.5,sigsize*3//8) *  (f2-f1) + (f1+f2)/2], axis=0)
    phase1 = np.cumsum(freqs1/fs)*2*np.pi
    sigs[:, 0] += np.sin(phase1)
    
    t_start = 0.
    name = 'spectrogram'
    view = ephyviewer.SpectrogramViewer.from_numpy( sigs, sample_rate, t_start, name, channel_names=None)
    
    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)
    win.add_view(view)
    
    
    # view2 = ephyviewer.TraceViewer.from_numpy(sigs, sample_rate, t_start, 'sigs', channel_names=None)
    # win.add_view(view2)

    if interactive:
        win.show()
        app.exec()
    else:
        # close thread properly
        win.close()


if __name__=='__main__':
    test_spectrogramviewer(interactive=True)



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)

from .myqt import QT
import pyqtgraph as pg
import numpy as np

from collections import OrderedDict

import time
import datetime


#TODO:
#  * xsize in navigation
#  * real time when possible

class NavigationToolBar(QT.QWidget) :
    """
    """
    time_changed = QT.pyqtSignal(float)
    xsize_changed = QT.pyqtSignal(float)
    auto_scale_requested = QT.pyqtSignal()

    def __init__(self, parent=None, show_play=True, show_step=True,
                                    show_scroll_time=True, show_spinbox=True,
                                    show_label_datetime=False, datetime0=None,
                                    datetime_format='%Y-%m-%d %H:%M:%S',
                                    show_global_xsize=True, show_auto_scale=True,
                                    play_interval = 0.1) :

        QT.QWidget.__init__(self, parent)

        self.setSizePolicy(QT.QSizePolicy.Minimum, QT.QSizePolicy.Maximum)

        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)



        #~ self.toolbar = QT.QToolBar()
        #~ self.mainlayout.addWidget(self.toolbar)
        #~ t = self.toolbar

        self.show_play = show_play
        self.show_step = show_step
        self.show_scroll_time = show_scroll_time
        self.show_spinbox = show_spinbox
        self.show_label_datetime = show_label_datetime
        self.show_global_xsize = show_global_xsize
        self.play_interval = play_interval

        self.datetime0 = datetime0
        self.datetime_format = datetime_format


        if show_scroll_time:
            #~ self.slider = QSlider()
            self.scroll_time = QT.QScrollBar(orientation=QT.Horizontal, minimum=0, maximum=1000)
            self.mainlayout.addWidget(self.scroll_time)
            self.scroll_time.valueChanged.connect(self.on_scroll_time_changed)

            #TODO min/max/step
            #~ self.scroll_time.valueChanged.disconnect(self.on_scroll_time_changed)
            #~ self.scroll_time.setValue(int(sr*t))
            #~ self.scroll_time.setPageStep(int(sr*self.xsize))
            #~ self.scroll_time.valueChanged.connect(self.on_scroll_time_changed)
            #~ self.scroll_time.setMinimum(0)
            #~ self.scroll_time.setMaximum(length)

        h = QT.QHBoxLayout()
        h.addStretch()
        self.mainlayout.addLayout(h)

        if show_play:
            but = QT.QPushButton(icon=QT.QIcon(':/media-playback-start.svg'))
            but.clicked.connect(self.on_play)
            h.addWidget(but)

            but = QT.QPushButton(icon=QT.QIcon(':/media-playback-stop.svg'))
            #~ but = QT.QPushButton(QT.QIcon(':/media-playback-stop.png'), '')
            but.clicked.connect(self.on_stop_pause)
            h.addWidget(but)

            h.addWidget(QT.QLabel('Speed:'))
            self.speedSpin = pg.SpinBox(bounds=(0.01, 100.), step=0.1, value=1.)
            if 'compactHeight' in self.speedSpin.opts:  # pyqtgraph >= 0.11.0
                self.speedSpin.setOpts(compactHeight=False)
            h.addWidget(self.speedSpin)
            self.speedSpin.valueChanged.connect(self.on_change_speed)

            self.speed = 1.

            #trick for separator
            h.addWidget(QT.QFrame(frameShape=QT.QFrame.VLine, frameShadow=QT.QFrame.Sunken))

            # add spacebar shortcut for play/pause
            play_pause_shortcut = QT.QShortcut(self)
            play_pause_shortcut.setKey(QT.QKeySequence(' '))
            play_pause_shortcut.activated.connect(self.on_play_pause_shortcut)

        self.steps = ['60 s', '10 s', '1 s', '100 ms', '50 ms', '5 ms', '1 ms', '200 us']

        if show_step:
            but = QT.QPushButton('<')
            but.clicked.connect(self.prev_step)
            h.addWidget(but)

            self.combo_step = QT.QComboBox()
            self.combo_step.addItems(self.steps)
            self.combo_step.setCurrentIndex(2)
            h.addWidget(self.combo_step)

            self.on_change_step(None)
            self.combo_step.currentIndexChanged.connect(self.on_change_step)

            but = QT.QPushButton('>')
            but.clicked.connect(self.next_step)
            h.addWidget(but)

            #trick for separator
            h.addWidget(QT.QFrame(frameShape=QT.QFrame.VLine, frameShadow=QT.QFrame.Sunken))

            # add shortcuts for stepping through time and changing step size
            shortcuts = [
                {'key': QT.Qt.Key_Left,  'callback': self.prev_step},
                {'key': QT.Qt.Key_Right, 'callback': self.next_step},
                {'key': QT.Qt.Key_Up,    'callback': self.increase_step},
                {'key': QT.Qt.Key_Down,  'callback': self.decrease_step},
                {'key': 'a',             'callback': self.prev_step},
                {'key': 'd',             'callback': self.next_step},
                {'key': 'w',             'callback': self.increase_step},
                {'key': 's',             'callback': self.decrease_step},
            ]
            for s in shortcuts:
                shortcut = QT.QShortcut(self)
                shortcut.setKey(QT.QKeySequence(s['key']))
                shortcut.activated.connect(s['callback'])


        if show_spinbox:
            h.addWidget(QT.QLabel('Time (s):'))
            self.spinbox_time =pg.SpinBox(decimals = 8, bounds = (-np.inf, np.inf),step = 0.05, siPrefix=False, suffix='', int=False)
            if 'compactHeight' in self.spinbox_time.opts:  # pyqtgraph >= 0.11.0
                self.spinbox_time.setOpts(compactHeight=False)
            h.addWidget(self.spinbox_time)
            #trick for separator
            h.addWidget(QT.QFrame(frameShape=QT.QFrame.VLine, frameShadow=QT.QFrame.Sunken))


            #~ h.addSeparator()
            self.spinbox_time.valueChanged.connect(self.on_spinbox_time_changed)

        if show_label_datetime:
            assert self.datetime0 is not None
            self.label_datetime = QT.QLabel('')
            h.addWidget(self.label_datetime)
            #trick for separator
            h.addWidget(QT.QFrame(frameShape=QT.QFrame.VLine, frameShadow=QT.QFrame.Sunken))

        if show_global_xsize:
            h.addWidget(QT.QLabel('Time width (s):'))
            self.spinbox_xsize =pg.SpinBox(value=3., decimals = 8, bounds = (0.001, np.inf),step = 0.1, siPrefix=False, suffix='', int=False)
            if 'compactHeight' in self.spinbox_xsize.opts:  # pyqtgraph >= 0.11.0
                self.spinbox_xsize.setOpts(compactHeight=False)
            h.addWidget(self.spinbox_xsize)
            #~ self.spinbox_xsize.valueChanged.connect(self.on_spinbox_xsize_changed)
            self.spinbox_xsize.valueChanged.connect(self.xsize_changed.emit)
            #trick for separator
            h.addWidget(QT.QFrame(frameShape=QT.QFrame.VLine, frameShadow=QT.QFrame.Sunken))

        if show_auto_scale:
            but = QT.PushButton('Auto scale')
            h.addWidget(but)
            but.clicked.connect(self.auto_scale_requested.emit)
            #~ h.addWidget(QT.QFrame(frameShape=QT.QFrame.VLine, frameShadow=QT.QFrame.Sunken))

        h.addStretch()

        self.timer_play = QT.QTimer(parent=self, interval=int(play_interval*1000))
        self.timer_play.timeout.connect(self.on_timer_play_interval)
        self.timer_delay = None

        # all in s

        self.t = 0 #  s
        self.set_start_stop(0., 0.1)

        self.last_time = None

    def on_play(self):
        self.last_time = time.time()
        self.timer_play.start()

    def on_stop_pause(self):
        self.timer_play.stop()
        self.seek(self.t)

    def on_play_pause_shortcut(self):
        if self.timer_play.isActive():
            self.on_stop_pause()
        else:
            self.on_play()

    def on_timer_play_interval(self):
        actual_time = time.time()
        #~ t = self.t +  self.play_interval*self.speed
        t = self.t +  (actual_time-self.last_time)*self.speed
        self.seek(t)
        self.last_time = actual_time

    def set_start_stop(self, t_start, t_stop, seek = True):
        #~ print 't_start', t_start, 't_stop', t_stop
        assert t_stop>t_start
        self.t_start = t_start
        self.t_stop = t_stop
        if seek:
            self.seek(self.t_start)
        if self.show_spinbox:
            self.spinbox_time.setMinimum(t_start)
            self.spinbox_time.setMaximum(t_stop)

    def on_change_step(self, val):
        text = str(self.combo_step.currentText ())

        if text.endswith('ms'):
            self.step_size = float(text[:-2])*1e-3
        elif text.endswith('us'):
            self.step_size = float(text[:-2])*1e-6
        else:
            self.step_size = float(text[:-1])
        #~ print('self.step_size', self.step_size)

    def prev_step(self):
        t = self.t -  self.step_size
        self.seek(t)

    def next_step(self):
        t = self.t +  self.step_size
        self.seek(t)

    def increase_step(self):
        new_index = max(self.combo_step.currentIndex()-1, 0)
        self.combo_step.setCurrentIndex(new_index)

    def decrease_step(self):
        new_index = min(self.combo_step.currentIndex()+1, self.combo_step.count()-1)
        self.combo_step.setCurrentIndex(new_index)

    def on_scroll_time_changed(self, pos):
        t = pos/1000.*(self.t_stop - self.t_start)+self.t_start
        self.seek(t, refresh_scroll = False)

    def on_spinbox_time_changed(self, val):
        self.seek(val, refresh_spinbox = False)

    #~ def on_spinbox_xsize_changed(self, val):
        #~ print('xsize', val)

    def seek(self , t, refresh_scroll = True, refresh_spinbox = True, emit=True):
        self.t = t
        if (self.t<self.t_start):
            self.t = self.t_start
        if (self.t>self.t_stop):
            self.t = self.t_stop
            if self.timer_play.isActive():
                self.timer_play.stop()
                #~ self.stop_pause()

        if refresh_scroll and self.show_scroll_time:
            self.scroll_time.valueChanged.disconnect(self.on_scroll_time_changed)
            pos = int((self.t - self.t_start)/(self.t_stop - self.t_start)*1000.)
            self.scroll_time.setValue(pos)
            self.scroll_time.valueChanged.connect(self.on_scroll_time_changed)

        if refresh_spinbox and self.show_spinbox:
            self.spinbox_time.valueChanged.disconnect(self.on_spinbox_time_changed)
            self.spinbox_time.setValue(t)
            self.spinbox_time.valueChanged.connect(self.on_spinbox_time_changed)

        if self.show_label_datetime:
            dt = self.datetime0 + datetime.timedelta(seconds=self.t)
            self.label_datetime.setText(dt.strftime(self.datetime_format))


        if emit:
            self.time_changed.emit(self.t)

    def on_change_speed(self , speed):
        self.speed = speed

    def set_settings(self, d):
        if hasattr(self, 'spinbox_xsize') and 'xsize' in d:
            self.spinbox_xsize.setValue(d['xsize'])

    def get_settings(self):
        d = {}
        if hasattr(self, 'spinbox_xsize'):
            d['xsize'] = float(self.spinbox_xsize.value())
        return d



import ephyviewer

from ephyviewer.base import ViewerBase






def test_mainviewer(interactive=False):

    class FakeView(ViewerBase):
        def __init__(self, name=''):
            ViewerBase.__init__(self, name)
            self.v = None
        def refresh(self):
            self.v = self.t
            #~ print('refresh', self.name, self.t)
            pass
        def set_settings(self, value):
            self.v = value
        def get_settings(self):
            return self.v

    app = ephyviewer.mkQApp()

    view1 = FakeView(name='view1')
    view2 = FakeView(name='view2')
    view3 = FakeView(name='view3')
    view4 = FakeView(name='view4')
    view5 = FakeView(name='view5')

    win = ephyviewer.MainViewer(settings_name='test0')
    win.add_view(view1)
    win.add_view(view2)

    win.add_view(view3, tabify_with='view2')
    win.add_view(view4, split_with='view1', orientation='horizontal')
    win.add_view(view5, location='bottom',  orientation='horizontal')

    if interactive:
        win.show()
        app.exec()
    else:
        # close thread properly
        win.close()

def test_mainviewer2(interactive=False):
    from  ephyviewer.tests.testing_tools import make_fake_video_source
    from  ephyviewer.tests.testing_tools import make_fake_signals
    from  ephyviewer.tests.testing_tools import make_fake_event_source
    from  ephyviewer.tests.testing_tools import make_fake_epoch_source


    app = ephyviewer.mkQApp()

    view1 = ephyviewer.TraceViewer(source=make_fake_signals(), name='signals')
    view2 = ephyviewer.VideoViewer(source=make_fake_video_source(), name='video')
    view3 = ephyviewer.EventList(source=make_fake_event_source(), name='events')
    view4 = ephyviewer.EpochViewer(source=make_fake_epoch_source(), name='epoch')
    view5 = ephyviewer.TimeFreqViewer(source=make_fake_signals(), name='timefreq')


    win = ephyviewer.MainViewer(debug=True, settings_name='test1', show_global_xsize=True, show_auto_scale=True)
    #TODO bug because new params!!!!!!!
    #~ win = ephyviewer.MainViewer(debug=True, show_global_xsize=True)
    win.add_view(view1)
    win.add_view(view5)
    win.add_view(view2)
    win.add_view(view4)
    win.add_view(view3, location='bottom',  orientation='horizontal')

    if interactive:
        win.show()
        app.exec()
    else:
        # close thread properly
        win.close()


def test_save_load_params(interactive=False):
    from  ephyviewer.tests.testing_tools import make_fake_signals


    app = ephyviewer.mkQApp()

    view1 = ephyviewer.TraceViewer(source=make_fake_signals(), name='signals')

    win = ephyviewer.MainViewer(debug=True, settings_name='test2', show_global_xsize=True, show_auto_scale=True)
    #~ print(win.settings_name)
    #~ exit()
    #TODO bug because new params!!!!!!!
    win.add_view(view1)

    if interactive:
        win.show()
        app.exec()
    else:
        # close thread properly
        win.close()



if __name__=='__main__':
    test_mainviewer(interactive=True)
    test_mainviewer2(interactive=True)
    test_save_load_params(interactive=True)



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)



class BaseDataSource:
    def __init__(self):
        pass

    @property
    def nb_channel(self):
        raise(NotImplementedError)

    def get_channel_name(self, chan=0):
        raise(NotImplementedError)

    @property
    def t_start(self):
        raise(NotImplementedError)

    @property
    def t_stop(self):
        raise(NotImplementedError)



from ephyviewer import QT, mkQApp


def test_icons(interactive=False):
    app = mkQApp()
    w = QT.QWidget()
    w.setWindowIcon(QT.QIcon(':/media-playback-start.svg'))

    if interactive:
    	w.show()
    	app.exec()


if __name__ == '__main__' :
	test_icons(interactive=True)




import sys
import os
import argparse
from ephyviewer.datasource import HAVE_NEO
from ephyviewer.standalone import all_neo_rawio_dict, rawio_gui_params
from ephyviewer import __version__

def launch_standalone_ephyviewer():
    from ephyviewer.standalone import WindowManager
    import pyqtgraph as pg
    assert HAVE_NEO, 'Must have Neo >= 0.6.0'
    import neo


    argv = sys.argv[1:]

    description = """
    Visualize electrophysiological data stored in files readable with Neo's
    RawIO classes
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('file_or_dir', default=None, nargs='?',
                        help='an optional path to a data file or directory, '
                             'which will be opened immediately (the file '
                             'format will be inferred from the file '
                             'extension, if possible; otherwise, --format is '
                             'required)')
    parser.add_argument('-V', '--version', action='version',
                        version='ephyviewer {}'.format(__version__))
    parser.add_argument('-f', '--format', default=None,
                        help='specify one of the following formats to '
                             'override the format detected automatically for '
                             'file_or_dir: {}'.format(
                             ', '.join(all_neo_rawio_dict.keys())))

    app = pg.mkQApp()

    manager = WindowManager()

    if len(argv)>=1:
        args = parser.parse_args(argv)
        file_or_dir_name = args.file_or_dir

        if args.format is None:
            #autoguess from extension
            neo_rawio_class = neo.rawio.get_rawio_class(file_or_dir_name)
        else:
            neo_rawio_class = all_neo_rawio_dict.get(args.format, None)

        assert neo_rawio_class is not None, 'Unknown format. Format list: {}'.format(', '.join(all_neo_rawio_dict.keys()))

        name = neo_rawio_class.__name__.replace('RawIO', '')
        if name in rawio_gui_params:
            raise(Exception('This IO requires additional parameters. Run ephyviewer without arguments to input these via the GUI.'))

        manager.load_dataset(neo_rawio_class=neo_rawio_class, file_or_dir_names=[file_or_dir_name])

    else:
        manager.open_dialog()

    if manager.windows:
        app.exec()

if __name__=='__main__':
    launch_standalone_ephyviewer()



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

from .myqt import QT
import pyqtgraph as pg

from .base import ViewerBase
from .datasource import InMemoryEventSource


def dataframe_on_qtable(qtable, df):
        qtable.clear()
        qtable.setColumnCount(len(df.columns))
        qtable.setRowCount(len(df.index))

        qtable.setHorizontalHeaderLabels(['{}'.format(c) for c in df.columns])
        qtable.setVerticalHeaderLabels(['{}'.format(c) for c in df.index])

        for r, row in enumerate(df.index):
            for c, col in enumerate(df.columns):
                try:
                    v = u'{}'.format(df.loc[row,col])
                    qtable.setItem(r, c,  QT.QTableWidgetItem(v))
                except Exception as e:
                    print('erreur', e)
                    print(r, row)
                    print(c, col)



class DataFrameView(ViewerBase):

    def __init__(self,  **kargs):
        ViewerBase.__init__(self, **kargs)

        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)


        self.qtable = QT.QTableWidget(selectionMode=QT.QAbstractItemView.SingleSelection,
                                                                            selectionBehavior=QT.QAbstractItemView.SelectRows)
        self.qtable.itemClicked.connect(self.on_selection_changed)
        self.mainlayout.addWidget(self.qtable)

        dataframe_on_qtable(self.qtable, self.source)

    def refresh(self):
        pass

    def on_selection_changed(self):
        if 'time' not in self.source.columns:
            return
        ind = [e.row() for e in  self.qtable.selectedIndexes() if e.column()==0]
        if len(ind)==1:
            t = self.source['time'].iloc[ind[0]]
            if t is not None:
                t = float(t)
                if not np.isnan(t):
                    self.time_changed.emit(t)



# -*- coding: utf-8 -*-
"""

Helper for importing Qt bindings library
see
http://mikeboers.com/blog/2015/07/04/static-libraries-in-a-dynamic-world#the-fold
"""


class ModuleProxy(object):

    def __init__(self, prefixes, modules):
        self.prefixes = prefixes
        self.modules = modules

    def __getattr__(self, name):
        for prefix in self.prefixes:
            fullname = prefix + name
            for module in self.modules:
                obj = getattr(module, fullname, None)
                if obj is not None:
                    setattr(self, name, obj) # cache it
                    return obj
        raise AttributeError(name)

QT_MODE = None


if QT_MODE is None:
    try:
        import PySide6
        from PySide6 import QtCore, QtGui, QtWidgets
        QT_MODE = 'PySide6'
    except ImportError:
        pass

#~ if QT_MODE is None:
    #~ try:
        #~ import PyQt6
        #~ from PyQt6 import QtCore, QtGui, QtWidgets
        #~ QT_MODE = 'PyQt6'
    #~ except ImportError:
        #~ pass

if QT_MODE is None:
    try:
        import PyQt5
        from PyQt5 import QtCore, QtGui, QtWidgets
        QT_MODE = 'PyQt5'
    except ImportError:
        pass

if QT_MODE is None:
    try:
        import PySide2
        from PySide2 import QtCore, QtGui, QtWidgets
        QT_MODE = 'PySide2'
    except ImportError:
        pass

if QT_MODE is None:
    try:
        import PyQt4
        from PyQt4 import QtCore, QtGui
        QT_MODE = 'PyQt4'
    except ImportError:
        pass


if QT_MODE is None:
    raise ImportError('Could not locate a supported Qt bindings library (PySide6, PyQt6, PyQt5, PySide2, PyQt4)')
# print(QT_MODE)

if QT_MODE == 'PyQt4':
    modules = [QtCore.Qt, QtCore, QtGui]
else:
    modules = [QtCore.Qt, QtCore, QtGui, QtWidgets]
QT = ModuleProxy(['', 'Q', 'Qt'], modules)

if QT_MODE.startswith('PySide'):
    QT.pyqtSignal = QtCore.Signal  # alias for cross-compatibility with PyQt

from pyqtgraph import mkQApp




# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

from .sourcebase import BaseDataSource



class BaseEventAndEpoch(BaseDataSource):
    type =None

    """

    Events is a list of either:
      * dict with 'time' and 'label'
      * np.array with dtype=[('time', 'float'), ('label, 'U')]

    """

    def __init__(self, all=[]):
        BaseDataSource.__init__(self)

        self.all = all

        s = [ np.min(e['time']) for e in self.all if len(e['time'])>0]
        self._t_start = min(s) if len(s)>0 else 0


    @property
    def nb_channel(self):
        return len(self.all)

    @property
    def t_start(self):
        return self._t_start

    @property
    def t_stop(self):
        return self._t_stop

    def get_channel_name(self, chan=0):
        return self.all[chan]['name']

    def get_size(self, chan=0):
        return self.all[chan]['time'].size






class InMemoryEventSource(BaseEventAndEpoch):
    type = 'Event'

    def __init__(self, all_events=[]):
        BaseEventAndEpoch.__init__(self, all=all_events)
        s = [ np.max(e['time']) for e in self.all  if len(e['time'])>0]
        self._t_stop = max(s) if len(s)>0 else 0


    def get_chunk(self, chan=0,  i_start=None, i_stop=None):
        ev_times = self.all[chan]['time'][i_start:i_stop]
        ev_labels = self.all[chan]['label'][i_start:i_stop]
        return ev_times, ev_labels

    def get_chunk_by_time(self, chan=0,  t_start=None, t_stop=None):
        ev_times = self.all[chan]['time']
        ev_labels = self.all[chan]['label']

        i1 = np.searchsorted(ev_times, t_start, side='left')
        i2 = np.searchsorted(ev_times, t_stop, side='left')
        sl = slice(i1, i2+1)
        return ev_times[sl], ev_labels[sl]



# -*- coding: utf-8 -*-
"""
This module include some facilities to make data source from:
  * neo.rawio so on disk sources (signals, spikes, epoch...)
  * neo objects in memory (neo.AnalogSignal, neo.Epoch, neo.SpikeTrain)

"""

from .sourcebase import BaseDataSource
import sys
import logging

import numpy as np

try:
    from distutils.version import LooseVersion as V
    import neo
    if V(neo.__version__)>='0.6.0':
        HAVE_NEO = True
        from neo.rawio.baserawio import BaseRawIO
    else:
        HAVE_NEO = False
        #~ print('neo version is too old', neo.__version__)
except ImportError:
    HAVE_NEO = False

from .signals import BaseAnalogSignalSource, InMemoryAnalogSignalSource
from .spikes import BaseSpikeSource, InMemorySpikeSource
from .events import BaseEventAndEpoch, InMemoryEventSource
from .epochs import InMemoryEpochSource



logger = logging.getLogger()

#~ print('HAVE_NEO', HAVE_NEO)



## neo.core stuff

class NeoAnalogSignalSource(InMemoryAnalogSignalSource):
    def __init__(self, neo_sig):
        signals = neo_sig.magnitude
        sample_rate = float(neo_sig.sampling_rate.rescale('Hz').magnitude)
        t_start = float(neo_sig.t_start.rescale('s').magnitude)

        InMemoryAnalogSignalSource.__init__(self, signals, sample_rate, t_start, channel_names=None)


class NeoSpikeTrainSource(InMemorySpikeSource):
    def __init__(self, neo_spiketrains=[]):
        all_spikes = []
        for neo_spiketrain in neo_spiketrains:
            name = neo_spiketrain.name
            if name is None:
                name = ''
            all_spikes.append({'time' : neo_spiketrain.times.rescale('s').magnitude,
                                        'name' : name})
        InMemorySpikeSource.__init__(self, all_spikes=all_spikes)

class NeoEventSource(InMemoryEventSource):
    def __init__(self, neo_events=[]):
        all_events = []
        for neo_event in neo_events:
            all_events.append({
                'name': neo_event.name,
                'time': neo_event.times.rescale('s').magnitude,
                'label': np.array(neo_event.labels),
            })
        InMemoryEventSource.__init__(self, all_events = all_events)

class NeoEpochSource(InMemoryEpochSource):
    def __init__(self, neo_epochs=[]):
        all_epochs = []
        for neo_epoch in neo_epochs:
            all_epochs.append({
                'name': neo_epoch.name,
                'time': neo_epoch.times.rescale('s').magnitude,
                'duration': neo_epoch.durations.rescale('s').magnitude,
                'label': np.array(neo_epoch.labels),
            })
        epoch_source = InMemoryEpochSource.__init__(self, all_epochs = all_epochs)





def get_sources_from_neo_segment(neo_seg):
    assert HAVE_NEO
    assert isinstance(neo_seg, neo.Segment)

    sources = {'signal':[], 'epoch':[], 'spike':[],'event':[],}

    for neo_sig in neo_seg.analogsignals:
        # normally neo signals are grouped by same sampling rate in one AnalogSignal
        # with shape (nb_channel, nb_sample)
        sources['signal'].append(NeoAnalogSignalSource(neo_sig))

    sources['spike'].append(NeoSpikeTrainSource(neo_seg.spiketrains))
    sources['event'].append(NeoEventSource(neo_seg.events))
    sources['epoch'].append(NeoEpochSource(neo_seg.epochs))


    return sources





## neo.rawio stuff

class AnalogSignalFromNeoRawIOSource(BaseAnalogSignalSource):
    def __init__(self, neorawio, channel_indexes=None, stream_index=None):
        """
        Create an analog signal source from a Neo RawIO.

        Parameters
        ----------
        neorawio : subclass of neo.rawio.BaseRawIO
            Neo RawIO reader from which to load signals.
        channel_indexes : list of ints
            Indexes of signals to use. Note that for Neo>=0.10, channels within
            a signal stream are indexed independently of channels in other
            streams; for Neo<0.10, channels are indexed globally, regardless of
            signal group membership. If None is passed, uses all channels within
            a stream (or all channels globally for Neo<0.10).
        stream_index : int
            Index of signal stream to use. If only one signal stream exists,
            this parameter is not required. For Neo<0.10, signal streams do not
            exist and this parameter must be None.
        """

        BaseAnalogSignalSource.__init__(self)
        self.with_scatter = False

        self.neorawio = neorawio
        if self.neorawio.header is None:
            self.neorawio.parse_header()

        if channel_indexes is None:
            channel_indexes = slice(None)
        self.channel_indexes = channel_indexes

        if V(neo.__version__)>='0.10.0':
            # Neo >= 0.10
            # - versions 0.10+ index channels within a stream
            if stream_index is not None:
                self.stream_index = stream_index
            elif self.neorawio.signal_streams_count() == 1:
                self.stream_index = 0
            else:
                raise ValueError(f'Because the Neo RawIO source contains multiple signal streams ({self.neorawio.signal_streams_count()}), stream_index must be provided')
            self.stream_id = self.neorawio.header['signal_streams'][self.stream_index]['id']
            signal_channels = self.neorawio.header['signal_channels']
            mask = signal_channels['stream_id'] == self.stream_id
            self.channels = signal_channels[mask][self.channel_indexes]
        else:
            # Neo < 0.10
            # - versions 0.6-0.9 index channels globally (ignoring signal group)
            assert stream_index is None, f'Neo version {neo.__version__} is installed, but only Neo>=0.10 uses stream_index'
            self.channels = self.neorawio.header['signal_channels'][self.channel_indexes]

        if V(neo.__version__)>='0.10.0':
            # Neo >= 0.10
            # - versions 0.10+ use stream_index as an argument often,
            #   but also require channel_indexes for get_chunk
            self.signal_indexing_kwarg = {'stream_index': self.stream_index}
            self.get_chunk_kwargs = {'stream_index': self.stream_index, 'channel_indexes': self.channel_indexes}
        else:
            # Neo < 0.10
            # - versions 0.6-0.9 use channel_indexes as an argument often
            self.signal_indexing_kwarg = {'channel_indexes': self.channel_indexes}
            self.get_chunk_kwargs = {'channel_indexes': self.channel_indexes}

        self.sample_rate = self.neorawio.get_signal_sampling_rate(**self.signal_indexing_kwarg)

        #TODO: something for multi segment
        self.block_index = 0
        self.seg_index = 0

    @property
    def nb_channel(self):
        return len(self.channels)

    def get_channel_name(self, chan=0):
        return self.channels[chan]['name']

    @property
    def t_start(self):
        t_start = self.neorawio.get_signal_t_start(self.block_index, self.seg_index,
                    **self.signal_indexing_kwarg)
        return t_start

    @property
    def t_stop(self):
        t_stop = self.t_start + self.get_length()/self.sample_rate
        return t_stop

    def get_length(self):
        length = self.neorawio.get_signal_size(self.block_index, self.seg_index,
                        **self.signal_indexing_kwarg)
        return length

    def get_gains(self):
        return self.channels['gain']

    def get_offsets(self):
        return self.channels['offset']

    def get_shape(self):
        return (self.get_length(), self.nb_channel)

    def get_chunk(self, i_start=None, i_stop=None):
        sigs = self.neorawio.get_analogsignal_chunk(block_index=self.block_index, seg_index=self.seg_index,
                        i_start=i_start, i_stop=i_stop, **self.get_chunk_kwargs)
        return sigs




class SpikeFromNeoRawIOSource(BaseSpikeSource):
    def __init__(self, neorawio, channel_indexes=None):
        self.neorawio =neorawio
        if channel_indexes is None:
            channel_indexes = slice(None)
        self.channel_indexes = channel_indexes

        if V(neo.__version__)>='0.10.0':
            # Neo >= 0.10
            # - versions 0.10+ have spike_channels
            self.channels = self.neorawio.header['spike_channels'][channel_indexes]
            self.get_chunk_kwarg = 'spike_channel_index'
        else:
            # Neo < 0.10
            # - versions 0.6-0.9 have unit_channels
            self.channels = self.neorawio.header['unit_channels'][channel_indexes]
            self.get_chunk_kwarg = 'unit_index'

        #TODO: something for multi segment
        self.block_index = 0
        self.seg_index = 0

    @property
    def nb_channel(self):
        return len(self.channels)

    def get_channel_name(self, chan=0):
        return self.channels[chan]['name']

    @property
    def t_start(self):
        t_start = self.neorawio.segment_t_start(self.block_index, self.seg_index)
        return t_start

    @property
    def t_stop(self):
        t_stop = self.neorawio.segment_t_stop(self.block_index, self.seg_index)
        return t_stop

    def get_chunk(self, chan=0,  i_start=None, i_stop=None):
        raise(NotImplementedError)

    def get_chunk_by_time(self, chan=0,  t_start=None, t_stop=None):
        spike_timestamp = self.neorawio.get_spike_timestamps(block_index=self.block_index,
                        seg_index=self.seg_index, **{self.get_chunk_kwarg: chan}, t_start=t_start, t_stop=t_stop)

        spike_times = self.neorawio.rescale_spike_timestamp(spike_timestamp, dtype='float64')

        return spike_times



class EpochFromNeoRawIOSource(BaseEventAndEpoch):
    def __init__(self, neorawio, channel_indexes=None):
        self.neorawio =neorawio
        if channel_indexes is None:
            channel_indexes = slice(None)
        self.channel_indexes = channel_indexes

        self.channels = self.neorawio.header['event_channels'][channel_indexes]

        #TODO: something for multi segment
        self.block_index = 0
        self.seg_index = 0

        self._cache_event = {}

    @property
    def nb_channel(self):
        return len(self.channels)

    def get_channel_name(self, chan=0):
        return self.channels[chan]['name']

    @property
    def t_start(self):
        t_start = self.neorawio.segment_t_start(self.block_index, self.seg_index)
        return t_start

    @property
    def t_stop(self):
        t_stop = self.neorawio.segment_t_stop(self.block_index, self.seg_index)
        return t_stop

    def get_chunk(self, chan=0,  i_start=None, i_stop=None):
        k = (self.block_index , self.seg_index, chan)
        if k not in self._cache_event:
            self._cache_event[k] = self.get_chunk_by_time(chan=chan,  t_start=None, t_stop=None)

        ep_times, ep_durations, ep_labels = self._cache_event[k]

        ep_times = ep_times[i_start:i_stop]
        if ep_durations is not None:
            ep_durations = ep_durations[i_start:i_stop]
        ep_labels = ep_labels[i_start:i_stop]

        return ep_times, ep_durations, ep_labels

    def get_chunk_by_time(self, chan=0,  t_start=None, t_stop=None):

        ep_timestamps, ep_durations, ep_labels = self.neorawio.get_event_timestamps(block_index=self.block_index,
                        seg_index=self.seg_index, event_channel_index=chan, t_start=t_start, t_stop=t_stop)

        ep_times = self.neorawio.rescale_event_timestamp(ep_timestamps, dtype='float64')

        if ep_durations is not None:
            ep_durations = self.neorawio.rescale_epoch_duration(ep_durations, dtype='float64')
        else:
            ep_durations = np.zeros_like(ep_times)

        return ep_times, ep_durations, ep_labels





def get_sources_from_neo_rawio(neorawio):
    assert HAVE_NEO
    assert isinstance(neorawio, BaseRawIO)

    if neorawio.header is None:
        logger.info('parse header')
        neorawio.parse_header()

    sources = {'signal':[], 'epoch':[], 'spike':[]}


    if hasattr(neorawio, 'signal_streams_count'):
        # Neo >= 0.10.0
        # - version 0.10 replaced signal groups with signal streams
        for stream_index in range(neorawio.signal_streams_count()):
            # one source per signal stream
            sources['signal'].append(AnalogSignalFromNeoRawIOSource(neorawio, stream_index=stream_index))
    elif hasattr(neorawio, 'get_group_signal_channel_indexes'):
        # Neo >= 0.9.0 and < 0.10
        # - version 0.9 renamed BaseRawIO.get_group_channel_indexes() to BaseRawIO.get_group_signal_channel_indexes()
        if neorawio.signal_channels_count() > 0:
            channel_indexes_list = neorawio.get_group_signal_channel_indexes()
            for channel_indexes in channel_indexes_list:
                # one source per channel group
                sources['signal'].append(AnalogSignalFromNeoRawIOSource(neorawio, channel_indexes=channel_indexes))
    elif hasattr(neorawio, 'get_group_channel_indexes'):
        # Neo < 0.9.0
        # - versions 0.6-0.8 have BaseRawIO.get_group_channel_indexes()
        if neorawio.signal_channels_count() > 0:
            channel_indexes_list = neorawio.get_group_channel_indexes()
            for channel_indexes in channel_indexes_list:
                # one source per channel group
                sources['signal'].append(AnalogSignalFromNeoRawIOSource(neorawio, channel_indexes=channel_indexes))


    if hasattr(neorawio, 'spike_channels_count'):
        # Neo >= 0.10
        # - version 0.10 renamed BaseRawIO.unit_channels_count() to BaseRawIO.spike_channels_count()
        if neorawio.spike_channels_count()>0:
            sources['spike'].append(SpikeFromNeoRawIOSource(neorawio, None))
    elif hasattr(neorawio, 'unit_channels_count'):
        # Neo < 0.10
        # - versions 0.6-0.9 have BaseRawIO.unit_channels_count()
        if neorawio.unit_channels_count()>0:
            sources['spike'].append(SpikeFromNeoRawIOSource(neorawio, None))


    if neorawio.event_channels_count()>0:
        sources['epoch'].append(EpochFromNeoRawIOSource(neorawio, None))



    return sources



# -*- coding: utf-8 -*-
from .version import version as __version__

#common tools
from .myqt import *
from .icons import *
from .datasource import *
from .mainviewer import MainViewer, compose_mainviewer_from_sources
from .navigation import NavigationToolBar



#Viewers
from .traceviewer import TraceViewer
from .videoviewer import VideoViewer
from .eventlist import EventList
from .epochviewer import EpochViewer
from .timefreqviewer import TimeFreqViewer
from .spectrogramviewer import SpectrogramViewer
from .dataframeview import DataFrameView
from .spiketrainviewer import SpikeTrainViewer


#Encoders
from .epochencoder import EpochEncoder



import ephyviewer


from  ephyviewer.tests.testing_tools import make_fake_event_source


def test_eventlist(interactive=False):
    source = make_fake_event_source()


    app = ephyviewer.mkQApp()
    view = ephyviewer.EventList(source=source, name='events')

    win = ephyviewer.MainViewer(debug=True)
    win.add_view(view)

    if interactive:
        win.show()
        app.exec()
    else:
        # close thread properly
        win.close()



if __name__=='__main__':
    test_eventlist(interactive=True)



import sys
from ephyviewer.myqt import QT_MODE

if QT_MODE == 'PySide6':
    from . import icons_PySide6 as icons
elif QT_MODE == 'PyQt6':
    # no more icons in PyQt6
    # https://stackoverflow.com/questions/66099225/how-can-resources-be-provided-in-pyqt6-which-has-no-pyrcc
    pass
elif QT_MODE == 'PyQt5':
    from . import icons_PyQt5 as icons
elif QT_MODE == 'PySide2':
    from . import icons_PySide2 as icons
elif QT_MODE == 'PyQt4':
    from . import  icons_PyQt4 as icons
else:
    raise ValueError('Could not load icons for unrecognized QT_MODE: ' + QT_MODE)



import ephyviewer
import ephyviewer.base



def test_base(interactive=False):
    app = ephyviewer.mkQApp()
    win = ephyviewer.base.ViewerBase()

    if interactive:
        win.show()
        app.exec()
    else:
        # close thread properly
        win.close()




if __name__=='__main__':
    test_base(interactive=True)



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)

from functools import lru_cache

from .myqt import QT
import pyqtgraph as pg


def get_dict_from_group_param(param, cascade = False):
    """
    Transform pyqtgraph params to dict.
    """
    assert param.type() == 'group'
    d = {}
    for p in param.children():
        if p.type() == 'group':
            if cascade:
                d[p.name()] = get_dict_from_group_param(p, cascade = True)
            continue
        else:
            d[p.name()] = p.value()
    return d

class ParamDialog(QT.QDialog):
    """
    Create a dialog with pyqtgraph params systrem.
    """
    def __init__(self,   params, title='', parent=None):
        QT.QDialog.__init__(self, parent = parent)

        self.setWindowTitle(title)
        self.setModal(True)

        self.params = pg.parametertree.Parameter.create( name=title, type='group', children = params)

        layout = QT.QVBoxLayout()
        self.setLayout(layout)

        self.tree_params = pg.parametertree.ParameterTree(parent  = self)
        self.tree_params.header().hide()
        self.tree_params.setParameters(self.params, showTop=True)
        #~ self.tree_params.setWindowFlags(QtCore.Qt.Window)
        layout.addWidget(self.tree_params)

        but = QT.QPushButton('OK')
        layout.addWidget(but)
        but.clicked.connect(self.accept)

    def get(self):
        return get_dict_from_group_param(self.params)




def create_plot_grid(graphiclayout, nb_column, visible_channels, ViewBoxClass=pg.ViewBox, vb_params={}):
    nb_channel = visible_channels.size
    nb_visible =sum(visible_channels)

    graphiclayout.clear()
    plots = [None] * nb_channel
    #~ images = [None] * nb_channel
    r,c = 0,0

    rowspan = nb_column
    colspan = nb_visible//nb_column
    graphiclayout.ci.currentRow = 0
    graphiclayout.ci.currentCol = 0
    for i in range(nb_channel):
        if not visible_channels[i]: continue

        viewBox = ViewBoxClass(**vb_params)
        #~ viewBox.setAspectLocked()
        plot = pg.PlotItem(viewBox=viewBox)
        plot.hideButtons()
        #~ plot.showAxis('left', False)
        #~ plot.showAxis('bottom', False)

        graphiclayout.addItem(plot, r, c)  # , rowspan, colspan)
        if r not in graphiclayout.ci.rows:
            graphiclayout.ci.rows[r] = {}
        graphiclayout.ci.rows[r][c] = plot
        graphiclayout.ci.items[plot] = [(r,c)]
        plots[i] = plot

        #~ images[i] = image = pg.ImageItem()
        #~ image.setPxMode(True)
        #~ plot.addItem(image)

        c+=1
        if c==nb_column:
            c=0
            r+=1

    return plots


@lru_cache(maxsize=None)
def mkCachedBrush(rgba):
    '''Create a QBrush from a color and cache it'''
    return pg.mkBrush(rgba)



# -*- coding: utf-8 -*-
#~ from __future__ import (unicode_literals, print_function, division, absolute_import)


import numpy as np

from .myqt import QT
import pyqtgraph as pg

from .base import ViewerBase
from .datasource import InMemoryEventSource, NeoEventSource




class EventList(ViewerBase):

    def __init__(self,  **kargs):
        ViewerBase.__init__(self, **kargs)

        self.mainlayout = QT.QVBoxLayout()
        self.setLayout(self.mainlayout)


        self.combo = QT.QComboBox()
        self.mainlayout.addWidget(self.combo)
        self.list_widget = QT.QListWidget()
        self.mainlayout.addWidget(self.list_widget)
        self.combo.currentIndexChanged.connect(self.refresh_list)

        self.combo.addItems([self.source.get_channel_name(i) for i in range(self.source.nb_channel) ])

        self.list_widget.itemClicked.connect(self.select_event)

    @classmethod
    def from_numpy(cls, all_events, name):
        source = InMemoryEventSource(all_events)
        view = cls(source=source, name=name)
        return view

    @classmethod
    def from_neo_events(cls, neo_events, name):
        source = NeoEventSource(neo_events)
        view = cls(source=source, name=name)
        return view

    def refresh(self):
        pass

    def refresh_list(self, ind):
        self.ind = ind
        self.list_widget.clear()
        #~ ev = self.source.all_events[ind]
        data = self.source.get_chunk(chan=ind,  i_start=None, i_stop=None)

        if len(data)==2:
            times, labels = data
        elif len(data)==3:
            times, _, labels = data
        elif len(data)==4:
            times, _, labels, _ = data
        else:
            raise ValueError("data has unexpected dimensions")

        for i in range(times.size):
            if labels is None:
                self.list_widget.addItem('{} : {:.3f}'.format(i, times[i]) )
            else:
                self.list_widget.addItem('{} : {:.3f} {}'.format(i, times[i], labels[i]) )


    def select_event(self):
        i = self.list_widget.currentRow()

        #~ ev = self.source.all_events[self.ind]
        #~ t = ev['time'][i]
        data = self.source.get_chunk(chan=self.ind,  i_start=i, i_stop=i+1)

        if len(data)==2:
            times, labels = data
        elif len(data)==3:
            times, _, labels = data
        elif len(data)==4:
            times, _, labels, _ = data
        else:
            raise ValueError("data has unexpected dimensions")

        if len(times)>0:
            t = float(times[0])
            self.time_changed.emit(t)



import ephyviewer



def test_navigation(interactive=False):
    app = ephyviewer.mkQApp()
    toolbar = ephyviewer.NavigationToolBar()

    if interactive:
        toolbar.show()
        app.exec()
    else:
        # close thread properly
        toolbar.close()


if __name__=='__main__':
    test_navigation(interactive=True)




import ephyviewer
import numpy as np
import os


def test_spikeinterface_viewer(interactive=False):
    import spikeinterface as si
    from spikeinterface.core.testing_tools import generate_recording, generate_sorting


    recording = generate_recording()
    sig_source = ephyviewer.SpikeInterfaceRecordingSource(recording=recording)

    sorting = generate_sorting()
    spike_source = ephyviewer.SpikeInterfaceSortingSource(sorting=sorting)

    app = ephyviewer.mkQApp()
    win = ephyviewer.MainViewer(debug=True, show_auto_scale=True)

    view = ephyviewer.TraceViewer(source=sig_source, name='signals')
    win.add_view(view)

    view = ephyviewer.SpikeTrainViewer(source=spike_source, name='spikes')
    win.add_view(view)


    if interactive:
        win.show()
        app.exec()
    else:
        # close thread properly
        win.close()



if __name__=='__main__':
    test_spikeinterface_viewer(interactive=True)
